Unit Modex;

Interface

{$G+}

Var
 Page:word;

Const
 MaxX=319;
 MaxY=199;
 NextLine=80;

Procedure GraphMode;
Procedure TextMode;

Procedure Clear;
Procedure Swap;

Procedure SetRGB(c,r,g,b:word);
Procedure SetPalette(Palette:pointer);

Procedure InitZBuffers;
Procedure PutZPixel(x,y,z,c:integer);

Procedure InitZGouraud(x,y,z,c:integer);
procedure ScanZGouraud(x,y,z,c:integer);
Procedure DrawZGouraud;

Procedure InitZPoly(x,y,z:integer);
Procedure ScanZPoly(x,y,z:integer);
Procedure DrawZPoly(c:integer);

Procedure InitPoly(x,y:integer);
Procedure ScanPoly(x,y:integer);
Procedure DrawPoly(c:integer);


Implementation

{.$define shift}

Const
{$ifdef shift}
 ColorShift=7;
 HalfColor=1 shl (ColorShift-1);
{$endif}

 PageSize=320*200 div 4;
 LastPage=3*PageSize;

Type
 pZBuffer=^TZBuffer;
 TZBuffer=array[0..319] of integer;
 TZBuffers=array[0..199] of pZBuffer;

Var
 ScanMin:array[0..199] of word;
 ScanMax:array[0..199] of word;
 ScanTop:word;
 ScanBottom:word;
 ScanX,ScanY:integer;

 ColorMin:array[0..199] of word;
 ColorMax:array[0..199] of word;
 ScanC:word;

 ScanMinZ:array[0..199] of integer;
 ScanMaxZ:array[0..199] of integer;
 ScanZ:integer;

 ZBuffers:TZBuffers;
 ZBuffer :pZBuffer;

function max(a,b:integer):integer; assembler;
 asm
  mov ax,a
  cmp ax,b
  jae @done
   mov ax,b
  @done:
 end;

Procedure GraphMode; assembler;
 asm
  mov ax,13h
  int 10h
{ Extended Mode X }
  mov dx,$3C4    { port[$3C4]:=4 }
  mov al,4
  out dx,al
  inc dx         { port[$3C5]:=(port[$3C5] and (not 8)) or 4 }
  in  al,dx
  and al,not 8
  or  al,4
  out dx,al

  mov dx,$3D4    { port[$3D4]:=$14 }
  mov al,$14
  out dx,al
  inc dx         { port[$3D5]:=(port[$3D5] and (not 64) }
  in  al,dx
  and al,not 64
  out dx,al

  dec dx         { port[$3D4]:=$17 }
  mov al,$17
  out dx,al
  inc dx
  in  al,dx      { port[$3D5]:=port[$3D5] or 64 }
  or  al,64
  out dx,al

{ clear screen }

  mov ax,$0F02    { port[$3C4]:=2 }
  mov dx,$3C4
  out dx,ax       { port[$3C5]:=1+2+3+4 }

  mov es,SegA000  { FillChar(Mem[$A000:0],(80*200),0) }
  xor di,di
  mov cx,80*200*4
  mov ax,0
  rep stosb

{ init swap page }

  mov Page,0 { PageSize }
 end;

Procedure TextMode; assembler;
 asm
  mov ax,03h
  int 10h
 end;

procedure Clear; assembler;
 asm
  mov ax,$0F02    { port[$3C4]:=2 }
  mov dx,$3C4
  out dx,ax       { port[$3C5]:=1+2+3+4 }

  mov es,SegA000
  mov di,Page
  mov cx,80*200
  mov ax,0
  rep stosb
 end;

procedure Swap; assembler;
 asm
  mov bx,Page
  mov dx,$3D4
  mov al,$C
  mov ah,bh
  out dx,ax
  inc ax
  mov ah,bl
  out dx,ax

  mov ax,Page
  sub ax,PageSize
  jns @ok
   mov ax,LastPage
  @ok:
  mov Page,ax
 end;

Procedure SetRGB(c,r,g,b:word); assembler;
 asm
  mov dx,$3C8
  mov ax,c
  out dx,al
  inc dx
  mov ax,r
  out dx,al
  mov ax,g
  out dx,al
  mov ax,b
  out dx,al
 end;

Procedure SetPalette(Palette:pointer); assembler;
 asm
 push ds
  mov dx,$3C8
  mov al,0
  out dx,al
  inc dx
  lds si,Palette
  mov cx,3*256
  rep outsb
 pop  ds
 end;


Procedure InitZBuffers; assembler;
 asm
  mov ax,32767
  mov bx,offset ZBuffers
  mov dx,200
  @nextbuffer:
  mov di,[bx]; add bx,2
  mov es,[bx]; add bx,2
  mov cx,320
  rep stosw
  dec dx
  jnz @nextbuffer
 end;

Procedure PutZPixel(x,y,z,c:integer);
 begin
  if ZBuffer^[x]>z then begin
   ZBuffer^[x]:=z;
   Port[$3c4]:=2;
   Port[$3c5]:=1 shl (x and 3);
   Mem[$A000:Page+80*y+x shr 2]:=c{$ifdef shift}shr ColorShift{$endif};
  end;
 end;


Procedure InitZGouraud(x,y,z,c:integer); assembler;
 asm
  mov ax,ds
  mov es,ax

  mov di,offset ScanMin
  mov ax,320
  mov cx,200
  rep stosw

  mov di,offset ScanMax
  mov ax,0
  mov cx,200
  rep stosw

  mov ax,x
  mov bx,y
  mov cx,c
  mov dx,z

 {$ifdef shift}
  shl cx,ColorShift
  add cx,HalfColor
 {$endif}

  mov ScanX,ax
  mov ScanY,bx
  mov ScanC,cx
  mov ScanZ,dx

  mov ScanTop,200
  mov ScanBottom,0
 end;

Procedure ScanZGouraud(x,y,z,c:integer);
 var
  dx,dy,dc,dz:integer;
  ix,iy,ic,iz:integer;
  ex,ey,ec,ez:integer;
  ee,ii:integer;
  ValidX,ValidY,Valid:boolean;
 begin
 {$ifdef shift}
  c:=c shl ColorShift;
 {$endif}
  dx:=x-ScanX; if dx<0 then begin dx:=-dx; ix:=-1 end else ix:=1;
  dy:=y-ScanY; if dy<0 then begin dy:=-dy; iy:=-1 end else iy:=1;
  dc:=c-ScanC; if dc<0 then begin dc:=-dc; ic:=-1 end else ic:=1;
  dz:=z-ScanZ; if dz<0 then begin dz:=-dz; iz:=-1 end else iz:=1;

  ex:=dx;
  ey:=dy;
  ec:=dc;
  ez:=dz;

  ee:=max(ex,max(ey,max(ec,ez)));

  if ScanY<ScanTop then if ScanY<0 then ScanTop:=0 else ScanTop:=ScanY;
  if ScanY>ScanBottom then if ScanY<200 then ScanBottom:=ScanY else ScanBottom:=199;

  ValidY:=(ScanY>=0)and(ScanY<200);
  ValidX:=(ScanX>=0)and(ScanX<320);
  Valid :=ValidX and ValidY;

  for ii:=0 to ee do begin

   if Valid then begin

    if ScanMin[ScanY]>ScanX then begin
      ScanMin[ScanY]:=ScanX;
     ColorMin[ScanY]:=ScanC;
     ScanMinZ[ScanY]:=ScanZ;
    end;

    if ScanMax[ScanY]<ScanX then begin
      ScanMax[ScanY]:=ScanX;
     ColorMax[ScanY]:=ScanC;
     ScanMaxZ[ScanY]:=ScanZ;
    end;

   end;

   if ey>ee then begin
    dec(ey,ee);
    inc(ScanY,iy);
    ValidY:=(ScanY>=0)and(ScanY<200);
    Valid :=ValidX and ValidY;
   end;
   inc(ey,dy);

   if ec>ee then begin
    dec(ec,ee);
    inc(ScanC,ic);
   end;
   inc(ec,dc);

   if ex>ee then begin
    dec(ex,ee);
    inc(ScanX,ix);
    ValidX:=(ScanX>=0)and(ScanX<320);
    Valid :=ValidX and ValidY;
   end;
   inc(ex,dx);

   if ez>ee then begin
    dec(ez,ee);
    inc(ScanZ,iz);
   end;
   inc(ez,dz);

  end;
 end;

Procedure DrawZGouraud;
 var
  y:word;
  x1,x2:word;
  c1,c2:integer;
  z1,z2:integer;
  dc,ic,ec:integer;
  dz,iz,ez:integer;
  dx,i ,ex:integer;
  f1,e,f2:integer;
  ee,ii:integer;
 begin
  For y:=ScanTop to ScanBottom do begin
   x1:=ScanMin[y];
   x2:=ScanMax[y];
   if x1<=x2 then begin

    ZBuffer:=ZBuffers[y];

    c1:=ColorMin[y];
    c2:=ColorMax[y];
    z1:=ScanMinZ[y];
    z2:=ScanMaxZ[y];

    dx:=x2-x1+1;
    dc:=c2-c1+1; if dc<0 then begin dc:=-dc; ic:=-1 end else ic:=+1;
    dz:=z2-z1+1; if dz<0 then begin dz:=-dz; iz:=-1 end else iz:=+1;

    ex:=dx;
    ec:=dc;
    ez:=dz;

    ee:=max(ex,max(ec,ez));

    for ii:=0 to ee do begin

     if ec>ee then begin
      dec(ec,ee);
      inc(c1,ic);
     end;
     inc(ec,dc);

     if ex>ee then begin
      PutZPixel(x1,y,z1,c1);
      dec(ex,ee);
      inc(x1);
     end;
     inc(ex,dx);

     if ez>ee then begin
      dec(ez,ee);
      inc(z1,iz);
     end;
     inc(ez,dz);

    end;
   end;
  end;

 end;

Procedure InitZPoly(x,y,z:integer); assembler;
 asm
  mov ax,ds
  mov es,ax

  mov di,offset ScanMin
  mov ax,320
  mov cx,200
  rep stosw

  mov di,offset ScanMax
  mov ax,0
  mov cx,200
  rep stosw

  mov ax,x
  mov bx,y
  mov dx,z

  mov ScanX,ax
  mov ScanY,bx
  mov ScanZ,dx

  mov ScanTop,200
  mov ScanBottom,0
 end;

Procedure ScanZPoly(x,y,z:integer);
 var
  dx,dy,dz:integer;
  ix,iy,iz:integer;
  ex,ey,ez:integer;
  ee,ii:integer;
  ValidX,ValidY,Valid:boolean;
 begin
  dx:=x-ScanX; if dx<0 then begin dx:=-dx; ix:=-1 end else ix:=1;
  dy:=y-ScanY; if dy<0 then begin dy:=-dy; iy:=-1 end else iy:=1;
  dz:=z-ScanZ; if dz<0 then begin dz:=-dz; iz:=-1 end else iz:=1;

  ex:=dx;
  ey:=dy;
  ez:=dz;

  ee:=max(ex,max(ey,ez));

  if ScanY<ScanTop then if ScanY<0 then ScanTop:=0 else ScanTop:=ScanY;
  if ScanY>ScanBottom then if ScanY<200 then ScanBottom:=ScanY else ScanBottom:=199;

  ValidY:=(ScanY>=0)and(ScanY<200);
  ValidX:=(ScanX>=0)and(ScanX<320);
  Valid :=ValidX and ValidY;

  for ii:=0 to ee do begin

   if Valid then begin

    if ScanMin[ScanY]>ScanX then begin
      ScanMin[ScanY]:=ScanX;
     ScanMinZ[ScanY]:=ScanZ;
    end;

    if ScanMax[ScanY]<ScanX then begin
      ScanMax[ScanY]:=ScanX;
     ScanMaxZ[ScanY]:=ScanZ;
    end;

   end;

   if ey>ee then begin
    dec(ey,ee);
    inc(ScanY,iy);
    ValidY:=(ScanY>=0)and(ScanY<200);
    Valid :=ValidX and ValidY;
   end;
   inc(ey,dy);

   if ex>ee then begin
    dec(ex,ee);
    inc(ScanX,ix);
    ValidX:=(ScanX>=0)and(ScanX<320);
    Valid :=ValidX and ValidY;
   end;
   inc(ex,dx);

   if ez>ee then begin
    dec(ez,ee);
    inc(ScanZ,iz);
   end;
   inc(ez,dz);

  end;
 end;

Procedure DrawZPoly(c:integer);
 var
  y:word;
  x1,x2:word;
  z1,z2:integer;
  dz,iz,ez:integer;
  dx,i ,ex:integer;
  f1,e,f2:integer;
  ee,ii:integer;
 begin
  For y:=ScanTop to ScanBottom do begin
   x1:=ScanMin[y];
   x2:=ScanMax[y];
   if x1<=x2 then begin

    ZBuffer:=ZBuffers[y];

    z1:=ScanMinZ[y];
    z2:=ScanMaxZ[y];

    dx:=x2-x1+1;
    dz:=z2-z1+1; if dz<0 then begin dz:=-dz; iz:=-1 end else iz:=+1;

    ex:=dx;
    ez:=dz;

    ee:=max(ex,ez);

    for ii:=0 to ee do begin

     if ex>ee then begin
      PutZPixel(x1,y,z1,c);
      dec(ex,ee);
      inc(x1);
     end;
     inc(ex,dx);

     if ez>ee then begin
      dec(ez,ee);
      inc(z1,iz);
     end;
     inc(ez,dz);

    end;
   end;
  end;

 end;

Procedure InitPoly(x,y:integer); assembler;
 asm
  mov ax,ds
  mov es,ax

  mov di,offset ScanMin
  mov ax,320
  mov cx,200
  rep stosw

  mov di,offset ScanMax
  mov ax,0
  mov cx,200
  rep stosw

  mov ax,x
  mov bx,y

  mov ScanX,ax
  mov ScanY,bx

  mov ScanTop,200
  mov ScanBottom,0
 end;

Procedure ScanPoly(x,y:integer);
 var
  dx,dy,dz:integer;
  ix,iy,iz:integer;
  ex,ey,ez:integer;
  ee,ii:integer;
  ValidX,ValidY,Valid:boolean;
 begin
  dx:=x-ScanX; if dx<0 then begin dx:=-dx; ix:=-1 end else ix:=1;
  dy:=y-ScanY; if dy<0 then begin dy:=-dy; iy:=-1 end else iy:=1;

  ValidY:=(ScanY>=0)and(ScanY<200);
  ValidX:=(ScanX>=0)and(ScanX<320);
  Valid :=ValidX and ValidY;

  if dx>dy then begin

   ee:=dy;
   for ii:=0 to dx do begin
    if Valid then begin
     if ScanX<ScanMin[ScanY] then ScanMin[ScanY]:=ScanX;
     if ScanX>ScanMax[ScanY] then ScanMax[ScanY]:=ScanX;
    end;
    inc(ee,dy);
    if ee>dx then begin
     dec(ee,dx);
     inc(ScanY,iy);
     ValidY:=(ScanY>=0)and(ScanY<200);
    end;
    inc(ScanX,ix);
    ValidX:=(ScanX>=0)and(ScanX<320);
    Valid :=ValidX and ValidY;
   end;

  end else begin

   ee:=dx;
   for ii:=0 to dy do begin
    if Valid then begin
     if ScanX<ScanMin[ScanY] then ScanMin[ScanY]:=ScanX;
     if ScanX>ScanMax[ScanY] then ScanMax[ScanY]:=ScanX;
    end;
    inc(ee,dx);
    if ee>dy then begin
     dec(ee,dy);
     inc(ScanX,ix);
     ValidX:=(ScanX>=0)and(ScanX<320);
    end;
    inc(ScanY,iy);
    ValidY:=(ScanY>=0)and(ScanY<200);
    Valid :=ValidX and ValidY;
   end;

  end;

 end;

Procedure DrawPoly(c:integer);
 var
  x,y:integer;
  x1,x2:integer;
 begin
  For y:=ScanTop to ScanBottom do begin
   x1:=ScanMin[y];
   x2:=ScanMax[y];
   if x1<=x2 then
    for x:=x1 to x2 do begin
     Port[$3c4]:=2;
     Port[$3c5]:=1 shl (x and 3);
     mem[SegA000:page+80*y+x shr 2]:=c;
    end;
  end;
 end;

var
 i:integer;
begin
 for i:=0 to 199 do new(ZBuffers[i]);
end.
